@@ -1,5 +1,6 @@
 // SPDX-License-Identifier: LGPL-2.1-only
-// Copyright (c) 2015 - 2017 DisplayLink (UK) Ltd.
+// Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
+
 #include <stddef.h>
 #include <stdint.h>
 #include <libdrm/drm.h>
@@ -25,10 +26,11 @@
 // ********************* Private part **************************
 
 #define MAX_DIRTS           16
+#define EVDI_INVALID_DEVICE_INDEX -1
 
 #define EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR 1
-#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 6
-#define EVDI_MODULE_COMPATIBILITY_VERSION_PATCHLEVEL 0
+#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 9
+#define EVDI_MODULE_COMPATIBILITY_VERSION_PATCH 0
 
 #define evdi_log(...) do {						\
 	if (g_evdi_logging.function) {					\
@@ -57,9 +59,52 @@ struct evdi_device_context {
 	int device_index;
 };
 
-static int do_ioctl(int fd, unsigned int request, void *data, const char *msg)
+#define EVDI_USAGE_LEN 64
+static evdi_handle card_usage[EVDI_USAGE_LEN];
+
+static int drm_ioctl(int fd, unsigned long request, void *arg)
 {
-	const int err = ioctl(fd, request, data);
+	int ret;
+
+	do {
+		ret = ioctl(fd, request, arg);
+	} while (ret == -1 && (errno == EINTR || errno == EAGAIN));
+
+	return ret;
+}
+
+static int drm_auth_magic(int fd, drm_magic_t magic)
+{
+	drm_auth_t auth;
+
+	memset(&auth, 0, sizeof(auth));
+	auth.magic = magic;
+	if (drm_ioctl(fd, DRM_IOCTL_AUTH_MAGIC, &auth))
+		return -errno;
+	return 0;
+}
+
+static int drm_is_master(int fd)
+{
+	/* Detect master by attempting something that requires master.
+	 *
+	 * Authenticating magic tokens requires master and 0 is an
+	 * internal kernel detail which we could use. Attempting this on
+	 * a master fd would fail therefore fail with EINVAL because 0
+	 * is invalid.
+	 *
+	 * A non-master fd will fail with EACCES, as the kernel checks
+	 * for master before attempting to do anything else.
+	 *
+	 * Since we don't want to leak implementation details, use
+	 * EACCES.
+	 */
+	return drm_auth_magic(fd, 0) != -EACCES;
+}
+
+static int do_ioctl(int fd, unsigned long request, void *data, const char *msg)
+{
+	const int err = drm_ioctl(fd, request, data);
 
 	if (err < 0)
 		evdi_log("Ioctl %s error: %s", msg, strerror(errno));
@@ -140,7 +185,7 @@ static int is_evdi_compatible(int fd)
 	evdi_log("LibEvdi version (%d.%d.%d)",
 		 LIBEVDI_VERSION_MAJOR,
 		 LIBEVDI_VERSION_MINOR,
-		 LIBEVDI_VERSION_PATCHLEVEL);
+		 LIBEVDI_VERSION_PATCH);
 
 	if (do_ioctl(fd, DRM_IOCTL_VERSION, &ver, "version") != 0)
 		return 0;
@@ -151,19 +196,15 @@ static int is_evdi_compatible(int fd)
 		 ver.version_patchlevel);
 
 	if (ver.version_major == EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR &&
-	    ver.version_minor == EVDI_MODULE_COMPATIBILITY_VERSION_MINOR)
+	    ver.version_minor >= EVDI_MODULE_COMPATIBILITY_VERSION_MINOR)
 		return 1;
 
 	evdi_log("Doesn't match LibEvdi compatibility one (%d.%d.%d)",
 		 EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR,
 		 EVDI_MODULE_COMPATIBILITY_VERSION_MINOR,
-		 EVDI_MODULE_COMPATIBILITY_VERSION_PATCHLEVEL);
+		 EVDI_MODULE_COMPATIBILITY_VERSION_PATCH);
 
-#ifdef NDEBUG
 	return 0;
-#else
-	return 1;
-#endif
 }
 
 static int is_evdi(int fd)
@@ -316,6 +357,39 @@ static void wait_for_master(const char *device_path)
 		evdi_log("Wait for master timed out");
 }
 
+static int open_as_slave(const char *device_path)
+{
+	int fd = 0;
+	int err = 0;
+
+	fd = open(device_path, O_RDWR);
+	if (fd < 0)
+		return -1;
+
+	if (drm_is_master(fd)) {
+		evdi_log("Process has master on %s, err: %s",
+			 device_path, strerror(errno));
+		err = drm_ioctl(fd, DRM_IOCTL_DROP_MASTER, NULL);
+	}
+
+	if (err < 0) {
+		evdi_log("Drop master on %s failed, err: %s",
+			 device_path, strerror(errno));
+		close(fd);
+		return err;
+	}
+
+	if (drm_is_master(fd)) {
+		evdi_log("Drop master on %s failed, err: %s",
+			 device_path, strerror(errno));
+		close(fd);
+		return -1;
+	}
+
+	evdi_log("Opened %s as slave drm device", device_path);
+	return fd;
+}
+
 static int wait_for_device(const char *device_path)
 {
 	const unsigned int TOTAL_WAIT_US = 5000000L;
@@ -325,7 +399,7 @@ static int wait_for_device(const char *device_path)
 
 	int fd = 0;
 
-	while ((fd = open(device_path, O_RDWR)) < 0 && cnt--)
+	while ((fd = open_as_slave(device_path)) < 0 && cnt--)
 		usleep(SLEEP_INTERVAL_US);
 
 	if (fd < 0)
@@ -347,7 +421,7 @@ static int open_device(int device)
 	fd = wait_for_device(dev);
 
 	if (fd >= 0) {
-		const int err = ioctl(fd, DRM_IOCTL_DROP_MASTER, NULL);
+		const int err = drm_ioctl(fd, DRM_IOCTL_DROP_MASTER, NULL);
 
 		if (err == 0)
 			evdi_log("Dropped master on %s", dev);
@@ -356,6 +430,146 @@ static int open_device(int device)
 	return fd;
 }
 
+
+static int write_add_device(const char *buffer, size_t buffer_length)
+{
+	FILE *add_devices = fopen("/sys/devices/evdi/add", "w");
+	int written = 0;
+
+	if (add_devices != NULL) {
+		written = fwrite(buffer,
+				 1,
+				 buffer_length,
+				 add_devices);
+		fclose(add_devices);
+	}
+
+	return written;
+}
+
+static int get_drm_device_index(const char *evdi_sysfs_drm_dir)
+{
+	struct dirent *fd_entry;
+	DIR *fd_dir;
+	int dev_index = EVDI_INVALID_DEVICE_INDEX;
+
+	fd_dir = opendir(evdi_sysfs_drm_dir);
+	if (fd_dir == NULL) {
+		evdi_log("Failed to open dir %s", evdi_sysfs_drm_dir);
+		return dev_index;
+	}
+
+	while ((fd_entry = readdir(fd_dir)) != NULL) {
+		if (strncmp(fd_entry->d_name, "card", 4) == 0)
+			dev_index = strtol(&fd_entry->d_name[4], NULL, 10);
+	}
+	closedir(fd_dir);
+
+	return dev_index;
+}
+
+static bool is_correct_parent_device(const char *dirname, const char *parent_device)
+{
+	char link_path[PATH_MAX];
+
+	snprintf(link_path, PATH_MAX - 7, "%s/device", dirname);
+
+	if (parent_device == NULL)
+		return access(link_path, F_OK) != 0;
+
+	char link_resolution[PATH_MAX];
+	const ssize_t link_resolution_len = readlink(link_path, link_resolution, PATH_MAX);
+
+	if (link_resolution_len == -1 || link_resolution_len == PATH_MAX)
+		return false;
+
+	link_resolution[link_resolution_len] = '\0';
+	char *parent_device_token = strrchr(link_resolution, '/');
+
+	if (strlen(parent_device) < 2)
+		return false;
+
+	parent_device_token++;
+	size_t len = strlen(parent_device_token);
+
+	bool is_same_device = strlen(parent_device) == len && strncmp(parent_device_token, parent_device, len) == 0;
+
+	return is_same_device;
+}
+
+static int find_unused_card_for(const char *parent_device)
+{
+	char evdi_platform_root[] = "/sys/bus/platform/devices";
+	struct dirent *fd_entry;
+	DIR *fd_dir;
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+
+	fd_dir = opendir(evdi_platform_root);
+	if (fd_dir == NULL) {
+		evdi_log("Failed to open dir %s", evdi_platform_root);
+		return device_index;
+	}
+
+	while ((fd_entry = readdir(fd_dir)) != NULL) {
+		if (strncmp(fd_entry->d_name, "evdi", 4) != 0)
+			continue;
+
+		char evdi_path[PATH_MAX];
+
+		snprintf(evdi_path, PATH_MAX, "%s/%s", evdi_platform_root, fd_entry->d_name);
+
+		if (!is_correct_parent_device(evdi_path, parent_device))
+			continue;
+
+		char evdi_drm_path[PATH_MAX];
+
+		snprintf(evdi_drm_path, PATH_MAX - strlen(evdi_path), "%s/drm", evdi_path);
+		int dev_index = get_drm_device_index(evdi_drm_path);
+
+		assert(dev_index < EVDI_USAGE_LEN && dev_index >= 0);
+
+		if (card_usage[dev_index] == EVDI_INVALID_HANDLE) {
+			device_index = dev_index;
+			break;
+		}
+	}
+	closedir(fd_dir);
+
+	return device_index;
+}
+
+static int get_generic_device(void)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+
+	device_index = find_unused_card_for(NULL);
+	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
+		evdi_log("Creating card in /sys/devices/platform");
+		write_add_device("1", 1);
+		device_index = find_unused_card_for(NULL);
+	}
+
+	return device_index;
+}
+
+static int get_device_attached_to_usb(const char *sysfs_parent_device)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
+	const char *parent_device = &sysfs_parent_device[4];
+
+	device_index = find_unused_card_for(parent_device);
+	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
+		evdi_log("Creating card for usb device %s in /sys/bus/platform/devices", parent_device);
+		const size_t len = strlen(sysfs_parent_device);
+
+		write_add_device(sysfs_parent_device, len);
+		device_index = find_unused_card_for(parent_device);
+	}
+
+	return device_index;
+}
+
+
 // ********************* Public part **************************
 
 evdi_handle evdi_open(int device)
@@ -370,6 +584,8 @@ evdi_handle evdi_open(int device)
 			if (h) {
 				h->fd = fd;
 				h->device_index = device;
+				card_usage[device] = h;
+				evdi_log("Using /dev/dri/card%d", device);
 			}
 		}
 		if (h == EVDI_INVALID_HANDLE)
@@ -378,17 +594,17 @@ evdi_handle evdi_open(int device)
 	return h;
 }
 
-enum evdi_device_status evdi_check_device(int device)
+static enum evdi_device_status evdi_device_to_platform(int device, char *path)
 {
 	struct dirent *fd_entry;
 	DIR *fd_dir;
 	enum evdi_device_status status = UNRECOGNIZED;
-	char path[PATH_MAX];
+	char card_path[PATH_MAX];
 
 	if (!device_exists(device))
 		return NOT_PRESENT;
 
-	fd_dir = opendir("/sys/devices/platform");
+	fd_dir = opendir("/sys/bus/platform/devices");
 	if (fd_dir == NULL) {
 		evdi_log("Failed to list platform devices");
 		return NOT_PRESENT;
@@ -399,10 +615,9 @@ enum evdi_device_status evdi_check_device(int device)
 			continue;
 
 		snprintf(path, PATH_MAX,
-			"/sys/devices/platform/%s/drm/card%d",
-			fd_entry->d_name,
-			device);
-		if (path_exists(path)) {
+			"/sys/bus/platform/devices/%s", fd_entry->d_name);
+		snprintf(card_path, PATH_MAX, "%s/drm/card%d", path, device);
+		if (path_exists(card_path)) {
 			status = AVAILABLE;
 			break;
 		}
@@ -412,30 +627,48 @@ enum evdi_device_status evdi_check_device(int device)
 	return status;
 }
 
+enum evdi_device_status evdi_check_device(int device)
+{
+	char path[PATH_MAX];
+
+	return evdi_device_to_platform(device, path);
+}
+
 int evdi_add_device(void)
 {
-	FILE *add_devices = fopen("/sys/devices/evdi/add", "w");
-	int written = 0;
+	return write_add_device("1", 1);
+}
 
-	if (add_devices != NULL) {
-		static const char devices_to_add[] = "1";
-		const size_t elem_bytes = 1;
+evdi_handle evdi_open_attached_to(const char *sysfs_parent_device)
+{
+	int device_index = EVDI_INVALID_DEVICE_INDEX;
 
-		written = fwrite(devices_to_add,
-				 elem_bytes,
-				 sizeof(devices_to_add),
-				 add_devices);
-		fclose(add_devices);
+	if (sysfs_parent_device == NULL)
+		device_index = get_generic_device();
+
+	if (sysfs_parent_device != NULL && strncmp(sysfs_parent_device, "usb:", 4) == 0 && strlen(sysfs_parent_device) > 4)
+		device_index = get_device_attached_to_usb(sysfs_parent_device);
+
+	if (device_index >= 0 && device_index < EVDI_USAGE_LEN)  {
+		evdi_handle handle = evdi_open(device_index);
+		return handle;
 	}
 
-	return written;
+	return EVDI_INVALID_HANDLE;
 }
 
+
 void evdi_close(evdi_handle handle)
 {
 	if (handle != EVDI_INVALID_HANDLE) {
 		close(handle->fd);
 		free(handle);
+		for (int device_index = 0; device_index < EVDI_USAGE_LEN; device_index++) {
+			if (card_usage[device_index] == handle) {
+				card_usage[device_index] = EVDI_INVALID_HANDLE;
+				evdi_log("Marking /dev/dri/card%d as unused", device_index);
+			}
+		}
 	}
 }
 
@@ -462,6 +695,19 @@ void evdi_disconnect(evdi_handle handle)
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "disconnect");
 }
 
+void evdi_enable_cursor_events(evdi_handle handle, bool enable)
+{
+	struct drm_evdi_enable_cursor_events cmd = {
+		.enable = enable,
+	};
+
+	evdi_log("%s cursor events on /dev/dri/card%d",
+		(enable ? "Enabling" : "Disabling"),
+		handle->device_index);
+
+	do_ioctl(handle->fd, DRM_IOCTL_EVDI_ENABLE_CURSOR_EVENTS, &cmd, "enable cursor events");
+}
+
 void evdi_grab_pixels(evdi_handle handle,
 		      struct evdi_rect *rects,
 		      int *num_rects)
@@ -553,6 +799,20 @@ bool evdi_request_update(evdi_handle handle, int bufferId)
 	}
 }
 
+void evdi_ddcci_response(evdi_handle handle, const unsigned char *buffer,
+			const uint32_t buffer_length,
+			const bool result)
+{
+	struct drm_evdi_ddcci_response cmd = {
+		.buffer = buffer,
+		.buffer_length = buffer_length,
+		.result = result,
+	};
+
+	do_ioctl(handle->fd, DRM_IOCTL_EVDI_DDCCI_RESPONSE, &cmd,
+		"ddcci_response");
+}
+
 static struct evdi_mode to_evdi_mode(struct drm_evdi_event_mode_changed *event)
 {
 	struct evdi_mode mode;
@@ -603,6 +863,8 @@ static struct evdi_cursor_set to_evdi_cursor_set(
 			cursor_set.buffer = malloc(size);
 			memcpy(cursor_set.buffer, ptr, size);
 			munmap(ptr, size);
+		} else {
+			evdi_log("Error: mmap failed with error: %s", strerror(errno));
 		}
 	}
 
@@ -620,6 +882,19 @@ static struct evdi_cursor_move to_evdi_cursor_move(
 	return cursor_move;
 }
 
+static struct evdi_ddcci_data to_evdi_ddcci_data(
+		struct drm_evdi_event_ddcci_data *event)
+{
+	struct evdi_ddcci_data ddcci_data;
+
+	ddcci_data.address = event->address;
+	ddcci_data.flags = event->flags;
+	ddcci_data.buffer = &event->buffer[0];
+	ddcci_data.buffer_length = event->buffer_length;
+
+	return ddcci_data;
+}
+
 static void evdi_handle_event(evdi_handle handle,
 			      struct evdi_event_context *evtctx,
 			      struct drm_event *e)
@@ -665,10 +940,13 @@ static void evdi_handle_event(evdi_handle handle,
 		if (evtctx->cursor_set_handler) {
 			struct drm_evdi_event_cursor_set *event =
 				(struct drm_evdi_event_cursor_set *) e;
+			struct evdi_cursor_set cursor_set = to_evdi_cursor_set(handle, event);
 
-			evtctx->cursor_set_handler(to_evdi_cursor_set(handle,
-								      event),
-						   evtctx->user_data);
+			if (cursor_set.enabled && cursor_set.buffer == NULL)
+				evdi_log("Error: Cursor buffer is null!");
+			else
+				evtctx->cursor_set_handler(cursor_set,
+							   evtctx->user_data);
 		}
 		break;
 
@@ -682,6 +960,16 @@ static void evdi_handle_event(evdi_handle handle,
 		}
 		break;
 
+	case DRM_EVDI_EVENT_DDCCI_DATA:
+		if (evtctx->ddcci_data_handler) {
+			struct drm_evdi_event_ddcci_data *event =
+				(struct drm_evdi_event_ddcci_data *) e;
+
+			evtctx->ddcci_data_handler(to_evdi_ddcci_data(event),
+						   evtctx->user_data);
+		}
+		break;
+
 	default:
 		evdi_log("Warning: Unhandled event");
 	}
@@ -718,7 +1006,7 @@ void evdi_get_lib_version(struct evdi_lib_version *version)
 	if (version != NULL) {
 		version->version_major = LIBEVDI_VERSION_MAJOR;
 		version->version_minor = LIBEVDI_VERSION_MINOR;
-		version->version_patchlevel = LIBEVDI_VERSION_PATCHLEVEL;
+		version->version_patchlevel = LIBEVDI_VERSION_PATCH;
 	}
 }
 
