@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: GPL-2.0-only
 /*
  * Copyright (C) 2012 Red Hat
- * Copyright (c) 2015 - 2017 DisplayLink (UK) Ltd.
+ * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
  *
  * Based on parts on udlfb.c:
  * Copyright (C) 2009 its respective authors
@@ -11,13 +11,16 @@
  * more details.
  */
 
-#include <drm/drmP.h>
+#include <linux/version.h>
 #include <drm/drm_crtc.h>
 #include <drm/drm_edid.h>
 #include <drm/drm_crtc_helper.h>
 #include <drm/drm_atomic_helper.h>
-#include <linux/version.h>
-#include "evdi_drv.h"
+#include "evdi_drm_drv.h"
+
+#if KERNEL_VERSION(5, 1, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#include <drm/drm_probe_helper.h>
+#endif
 
 /*
  * dummy connector to just get EDID,
@@ -33,7 +36,7 @@ static int evdi_get_modes(struct drm_connector *connector)
 	edid = (struct edid *)evdi_painter_get_edid_copy(evdi);
 
 	if (!edid) {
-#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE
+#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE || defined(EL8)
 		drm_connector_update_edid_property(connector, NULL);
 #else
 		drm_mode_connector_update_edid_property(connector, NULL);
@@ -41,17 +44,20 @@ static int evdi_get_modes(struct drm_connector *connector)
 		return 0;
 	}
 
-#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE
+#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE || defined(EL8)
 	ret = drm_connector_update_edid_property(connector, edid);
 #else
 	ret = drm_mode_connector_update_edid_property(connector, edid);
 #endif
 
-	if (!ret)
-		drm_add_edid_modes(connector, edid);
-	else
-		EVDI_ERROR("Failed to set edid modes! error: %d", ret);
+	if (ret) {
+		EVDI_ERROR("Failed to set edid property! error: %d", ret);
+		goto err;
+	}
 
+	ret = drm_add_edid_modes(connector, edid);
+	EVDI_INFO("(card%d) Edid property set", evdi->dev_index);
+err:
 	kfree(edid);
 	return ret;
 }
@@ -66,7 +72,7 @@ static enum drm_mode_status evdi_mode_valid(struct drm_connector *connector,
 		return MODE_OK;
 
 	if (mode_area > evdi->sku_area_limit) {
-		EVDI_WARN("(dev=%d) Mode %dx%d@%d rejected\n",
+		EVDI_WARN("(card%d) Mode %dx%d@%d rejected\n",
 			evdi->dev_index,
 			mode->hdisplay,
 			mode->vdisplay,
@@ -83,12 +89,12 @@ evdi_detect(struct drm_connector *connector, __always_unused bool force)
 	struct evdi_device *evdi = connector->dev->dev_private;
 
 	EVDI_CHECKPT();
-	if (evdi_painter_is_connected(evdi)) {
-		EVDI_DEBUG("(dev=%d) poll connector state: connected\n",
+	if (evdi_painter_is_connected(evdi->painter)) {
+		EVDI_INFO("(card%d) Connector state: connected\n",
 			   evdi->dev_index);
 		return connector_status_connected;
 	}
-	EVDI_DEBUG("(dev=%d) poll connector state: disconnected\n",
+	EVDI_VERBOSE("(card%d) Connector state: disconnected\n",
 		   evdi->dev_index);
 	return connector_status_disconnected;
 }
@@ -102,12 +108,19 @@ static void evdi_connector_destroy(struct drm_connector *connector)
 
 static struct drm_encoder *evdi_best_encoder(struct drm_connector *connector)
 {
+#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	struct drm_encoder *encoder;
 
+	drm_connector_for_each_possible_encoder(connector, encoder) {
+		return encoder;
+	}
+
+	return NULL;
+#else
 	return drm_encoder_find(connector->dev,
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
 				NULL,
-#endif
 				connector->encoder_ids[0]);
+#endif
 }
 
 static struct drm_connector_helper_funcs evdi_connector_helper_funcs = {
@@ -117,15 +130,9 @@ static struct drm_connector_helper_funcs evdi_connector_helper_funcs = {
 };
 
 static const struct drm_connector_funcs evdi_connector_funcs = {
-#if KERNEL_VERSION(4, 14, 0) > LINUX_VERSION_CODE
-	.dpms = drm_atomic_helper_connector_dpms,
-#endif
 	.detect = evdi_detect,
 	.fill_modes = drm_helper_probe_single_connector_modes,
 	.destroy = evdi_connector_destroy,
-#if KERNEL_VERSION(4, 14, 0) > LINUX_VERSION_CODE
-	.set_property = drm_atomic_helper_connector_set_property,
-#endif
 	.reset = drm_atomic_helper_connector_reset,
 	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
 	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state
@@ -134,6 +141,7 @@ static const struct drm_connector_funcs evdi_connector_funcs = {
 int evdi_connector_init(struct drm_device *dev, struct drm_encoder *encoder)
 {
 	struct drm_connector *connector;
+	struct evdi_device *evdi = dev->dev_private;
 
 	connector = kzalloc(sizeof(struct drm_connector), GFP_KERNEL);
 	if (!connector)
@@ -147,16 +155,12 @@ int evdi_connector_init(struct drm_device *dev, struct drm_encoder *encoder)
 
 	drm_connector_register(connector);
 
-#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE
+	evdi->conn = connector;
+
+#if KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE  || defined(EL8)
 	drm_connector_attach_encoder(connector, encoder);
 #else
 	drm_mode_connector_attach_encoder(connector, encoder);
 #endif
-
-#if KERNEL_VERSION(4, 9, 0) > LINUX_VERSION_CODE
-	drm_object_attach_property(&connector->base,
-				   dev->mode_config.dirty_info_property, 1);
-#endif
-
 	return 0;
 }
