@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: GPL-2.0-only
 /*
  * Copyright (C) 2012 Red Hat
- * Copyright (c) 2015 - 2016 DisplayLink (UK) Ltd.
+ * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
  *
  * Based on parts on udlfb.c:
  * Copyright (C) 2009 its respective authors
@@ -17,23 +17,19 @@
 #endif /* CONFIG_FB */
 #include <linux/dma-buf.h>
 #include <linux/version.h>
+#if KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#else
 #include <drm/drmP.h>
+#endif
 #include <drm/drm_crtc.h>
 #include <drm/drm_crtc_helper.h>
 #include <drm/drm_fb_helper.h>
 #include <drm/drm_atomic.h>
-#include "evdi_drv.h"
-
-#if KERNEL_VERSION(4, 12, 0) > LINUX_VERSION_CODE
-static inline void drm_gem_object_put(struct drm_gem_object *obj)
-{
-	drm_gem_object_unreference(obj);
-}
-static inline void drm_gem_object_put_unlocked(struct drm_gem_object *obj)
-{
-	drm_gem_object_unreference_unlocked(obj);
-}
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_damage_helper.h>
 #endif
+#include "evdi_drm_drv.h"
+
 
 struct evdi_fbdev {
 	struct drm_fb_helper helper;
@@ -89,6 +85,7 @@ struct drm_clip_rect evdi_framebuffer_sanitize_rect(
 	return rect;
 }
 
+#ifdef CONFIG_FB
 static int evdi_handle_damage(struct evdi_framebuffer *fb,
 		       int x, int y, int width, int height)
 {
@@ -102,13 +99,12 @@ static int evdi_handle_damage(struct evdi_framebuffer *fb,
 
 	if (!fb->active)
 		return 0;
-	evdi_painter_set_scanout_buffer(evdi, fb);
+	evdi_painter_set_scanout_buffer(evdi->painter, fb);
 	evdi_painter_mark_dirty(evdi, &rect);
 
 	return 0;
 }
 
-#ifdef CONFIG_FB
 static int evdi_fb_mmap(struct fb_info *info, struct vm_area_struct *vma)
 {
 	unsigned long start = vma->vm_start;
@@ -225,6 +221,8 @@ static struct fb_ops evdifb_ops = {
 };
 #endif /* CONFIG_FB */
 
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
+#else
 /*
  * Function taken from
  * https://lists.freedesktop.org/archives/dri-devel/2018-September/188716.html
@@ -245,19 +243,16 @@ static int evdi_user_framebuffer_dirty(
 	struct drm_atomic_state *state;
 	struct drm_plane *plane;
 	int ret = 0;
-	int i;
+	unsigned int i;
 
 	EVDI_CHECKPT();
 
 	drm_modeset_acquire_init(&ctx,
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
 		/*
 		 * When called from ioctl, we are interruptable,
 		 * but not when called internally (ie. defio worker)
 		 */
-		file_priv ? DRM_MODESET_ACQUIRE_INTERRUPTIBLE :
-#endif
-		0);
+		file_priv ? DRM_MODESET_ACQUIRE_INTERRUPTIBLE :	0);
 
 	state = drm_atomic_state_alloc(fb->dev);
 	if (!state) {
@@ -269,15 +264,9 @@ static int evdi_user_framebuffer_dirty(
 	for (i = 0; i < num_clips; ++i)
 		evdi_painter_mark_dirty(evdi, &clips[i]);
 
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
 retry:
-#endif
 
-#if KERNEL_VERSION(4, 3, 0) <= LINUX_VERSION_CODE
 	drm_for_each_plane(plane, fb->dev) {
-#else
-	list_for_each_entry(plane, &fb->dev->mode_config.plane_list, head) {
-#endif
 		struct drm_plane_state *plane_state;
 
 		if (plane->state->fb != fb)
@@ -300,26 +289,20 @@ retry:
 out:
 	if (ret == -EDEADLK) {
 		drm_atomic_state_clear(state);
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
 		ret = drm_modeset_backoff(&ctx);
 		if (!ret)
 			goto retry;
-#else
-		drm_modeset_backoff(&ctx);
-#endif
 	}
 
-#if KERNEL_VERSION(4, 10, 0) <= LINUX_VERSION_CODE
-	drm_atomic_state_put(state);
-#else
-	drm_atomic_state_free(state);
-#endif
+	if (state)
+		drm_atomic_state_put(state);
 
 	drm_modeset_drop_locks(&ctx);
 	drm_modeset_acquire_fini(&ctx);
 
 	return ret;
 }
+#endif
 
 static int evdi_user_framebuffer_create_handle(struct drm_framebuffer *fb,
 					       struct drm_file *file_priv,
@@ -336,8 +319,11 @@ static void evdi_user_framebuffer_destroy(struct drm_framebuffer *fb)
 
 	EVDI_CHECKPT();
 	if (efb->obj)
+#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE || defined(EL8)
+		drm_gem_object_put(&efb->obj->base);
+#else
 		drm_gem_object_put_unlocked(&efb->obj->base);
-
+#endif
 	drm_framebuffer_cleanup(fb);
 	kfree(efb);
 }
@@ -345,25 +331,21 @@ static void evdi_user_framebuffer_destroy(struct drm_framebuffer *fb)
 static const struct drm_framebuffer_funcs evdifb_funcs = {
 	.create_handle = evdi_user_framebuffer_create_handle,
 	.destroy = evdi_user_framebuffer_destroy,
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
+	.dirty = drm_atomic_helper_dirtyfb,
+#else
 	.dirty = evdi_user_framebuffer_dirty,
+#endif
 };
 
 static int
 evdi_framebuffer_init(struct drm_device *dev,
 		      struct evdi_framebuffer *efb,
-#if KERNEL_VERSION(4, 5, 0) > LINUX_VERSION_CODE
-		      struct drm_mode_fb_cmd2 *mode_cmd,
-#else
 		      const struct drm_mode_fb_cmd2 *mode_cmd,
-#endif
 		      struct evdi_gem_object *obj)
 {
 	efb->obj = obj;
-#if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
-	drm_helper_mode_fill_fb_struct(&efb->base, mode_cmd);
-#else
 	drm_helper_mode_fill_fb_struct(dev, &efb->base, mode_cmd);
-#endif
 	return drm_framebuffer_init(dev, &efb->base, &evdifb_funcs);
 }
 
@@ -431,21 +413,22 @@ static int evdifb_create(struct drm_fb_helper *helper,
 	info->fix.smem_len = size;
 	info->fix.smem_start = (unsigned long)efbdev->efb.obj->vmapping;
 
-#if KERNEL_VERSION(4, 20, 0) > LINUX_VERSION_CODE
-	info->flags = FBINFO_DEFAULT | FBINFO_CAN_FORCE_OUTPUT;
-#else
+#if KERNEL_VERSION(4, 20, 0) <= LINUX_VERSION_CODE || defined(EL8)
 	info->flags = FBINFO_DEFAULT;
+#else
+	info->flags = FBINFO_DEFAULT | FBINFO_CAN_FORCE_OUTPUT;
 #endif
 
 	efbdev->fb_ops = evdifb_ops;
 	info->fbops = &efbdev->fb_ops;
-#if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
-	drm_fb_helper_fill_fix(info, fb->pitches[0], fb->depth);
+
+#if KERNEL_VERSION(5, 2, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_fb_helper_fill_info(info, &efbdev->helper, sizes);
 #else
 	drm_fb_helper_fill_fix(info, fb->pitches[0], fb->format->depth);
-#endif
 	drm_fb_helper_fill_var(info, &efbdev->helper, sizes->fb_width,
 			       sizes->fb_height);
+#endif
 
 	ret = fb_alloc_cmap(&info->cmap, 256, 0);
 	if (ret) {
@@ -458,7 +441,11 @@ static int evdifb_create(struct drm_fb_helper *helper,
 
 	return ret;
  out_gfree:
+#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_gem_object_put(&efbdev->efb.obj->base);
+#else
 	drm_gem_object_put_unlocked(&efbdev->efb.obj->base);
+#endif
  out:
 	return ret;
 }
@@ -481,9 +468,15 @@ static void evdi_fbdev_destroy(__always_unused struct drm_device *dev,
 		framebuffer_release(info);
 	}
 	drm_fb_helper_fini(&efbdev->helper);
-	drm_framebuffer_unregister_private(&efbdev->efb.base);
-	drm_framebuffer_cleanup(&efbdev->efb.base);
-	drm_gem_object_put_unlocked(&efbdev->efb.obj->base);
+	if (efbdev->efb.obj) {
+		drm_framebuffer_unregister_private(&efbdev->efb.base);
+		drm_framebuffer_cleanup(&efbdev->efb.base);
+#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE || defined(EL8)
+		drm_gem_object_put(&efbdev->efb.obj->base);
+#else
+		drm_gem_object_put_unlocked(&efbdev->efb.obj->base);
+#endif
+	}
 }
 
 int evdi_fbdev_init(struct drm_device *dev)
@@ -500,8 +493,8 @@ int evdi_fbdev_init(struct drm_device *dev)
 	evdi->fbdev = efbdev;
 	drm_fb_helper_prepare(dev, &efbdev->helper, &evdi_fb_helper_funcs);
 
-#if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
-	ret = drm_fb_helper_init(dev, &efbdev->helper, 1, 1);
+#if KERNEL_VERSION(5, 7, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	ret = drm_fb_helper_init(dev, &efbdev->helper);
 #else
 	ret = drm_fb_helper_init(dev, &efbdev->helper, 1);
 #endif
@@ -510,7 +503,10 @@ int evdi_fbdev_init(struct drm_device *dev)
 		return ret;
 	}
 
+#if KERNEL_VERSION(5, 7, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#else
 	drm_fb_helper_single_add_all_connectors(&efbdev->helper);
+#endif
 
 	ret = drm_fb_helper_initial_config(&efbdev->helper, 32);
 	if (ret) {
@@ -545,36 +541,28 @@ void evdi_fbdev_unplug(struct drm_device *dev)
 		struct fb_info *info;
 
 		info = efbdev->helper.fbdev;
+#if KERNEL_VERSION(5, 6, 0) <= LINUX_VERSION_CODE
+		unregister_framebuffer(info);
+#else
 		unlink_framebuffer(info);
+#endif
 	}
 }
 #endif /* CONFIG_FB */
 
 int evdi_fb_get_bpp(uint32_t format)
 {
-#if KERNEL_VERSION(4, 10, 0) <= LINUX_VERSION_CODE
 	const struct drm_format_info *info = drm_format_info(format);
 
 	if (!info)
 		return 0;
 	return info->cpp[0] * 8;
-#else
-	unsigned int depth;
-	int bpp;
-
-	drm_fb_get_bpp_depth(format, &depth, &bpp);
-	return bpp;
-#endif
 }
 
 struct drm_framebuffer *evdi_fb_user_fb_create(
 					struct drm_device *dev,
 					struct drm_file *file,
-#if KERNEL_VERSION(4, 5, 0) > LINUX_VERSION_CODE
-					struct drm_mode_fb_cmd2 *mode_cmd)
-#else
 					const struct drm_mode_fb_cmd2 *mode_cmd)
-#endif
 {
 	struct drm_gem_object *obj;
 	struct evdi_framebuffer *efb;
@@ -587,27 +575,24 @@ struct drm_framebuffer *evdi_fb_user_fb_create(
 		return ERR_PTR(-EINVAL);
 	}
 
-#if KERNEL_VERSION(4, 7, 0) > LINUX_VERSION_CODE
-	obj = drm_gem_object_lookup(dev, file, mode_cmd->handles[0]);
-#else
 	obj = drm_gem_object_lookup(file, mode_cmd->handles[0]);
-#endif
 	if (obj == NULL)
 		return ERR_PTR(-ENOENT);
 
-	size = mode_cmd->pitches[0] * mode_cmd->height;
+	size = mode_cmd->offsets[0] + mode_cmd->pitches[0] * mode_cmd->height;
 	size = ALIGN(size, PAGE_SIZE);
 
 	if (size > obj->size) {
-		DRM_ERROR("object size not sufficient for fb %d %zu %d %d\n",
-			  size, obj->size, mode_cmd->pitches[0],
-			  mode_cmd->height);
+		DRM_ERROR("object size not sufficient for fb %d %zu %u %d %d\n",
+			  size, obj->size, mode_cmd->offsets[0],
+			  mode_cmd->pitches[0], mode_cmd->height);
 		goto err_no_mem;
 	}
 
 	efb = kzalloc(sizeof(*efb), GFP_KERNEL);
 	if (efb == NULL)
 		goto err_no_mem;
+	efb->base.obj[0] = obj;
 
 	ret = evdi_framebuffer_init(dev, efb, mode_cmd, to_evdi_bo(obj));
 	if (ret)
