@@ -3,7 +3,7 @@
  * evdi_ioc32.c
  *
  * Copyright (c) 2016 The Chromium OS Authors
- * Copyright (c) 2017 DisplayLink (UK) Ltd.
+ * Copyright (c) 2017 - 2020 DisplayLink (UK) Ltd.
  *
  * This program is free software; you can redistribute  it and/or modify it
  * under  the terms of  the GNU General  Public License as published by the
@@ -21,11 +21,17 @@
 
 #include <linux/compat.h>
 
+#include <linux/version.h>
+#if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_ioctl.h>
+#elif KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#else
 #include <drm/drmP.h>
+#endif
 #include <drm/drm_edid.h>
 #include "evdi_drm.h"
 
-#include "evdi_drv.h"
+#include "evdi_drm_drv.h"
 
 struct drm_evdi_connect32 {
 	int32_t connected;
@@ -55,11 +61,15 @@ static int compat_evdi_connect(struct file *file,
 	if (copy_from_user(&req32, (void __user *)arg, sizeof(req32)))
 		return -EFAULT;
 
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE && KERNEL_VERSION(5, 14, 0) >= LINUX_VERSION_CODE
 	request = compat_alloc_user_space(sizeof(*request));
-#if KERNEL_VERSION(5, 0, 0) > LINUX_VERSION_CODE
-	if (!access_ok(VERIFY_WRITE, request, sizeof(*request))
 #else
+	request = kmalloc(sizeof(*request), GFP_USER);
+#endif
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
 	if (!access_ok(request, sizeof(*request))
+#else
+	if (!access_ok(VERIFY_WRITE, request, sizeof(*request))
 #endif
 		|| __put_user(req32.connected, &request->connected)
 		|| __put_user(req32.dev_index, &request->dev_index)
@@ -83,11 +93,15 @@ static int compat_evdi_grabpix(struct file *file,
 	if (copy_from_user(&req32, (void __user *)arg, sizeof(req32)))
 		return -EFAULT;
 
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE && KERNEL_VERSION(5, 14, 0) >= LINUX_VERSION_CODE
 	request = compat_alloc_user_space(sizeof(*request));
-#if KERNEL_VERSION(5, 0, 0) > LINUX_VERSION_CODE
-	if (!access_ok(VERIFY_WRITE, request, sizeof(*request))
 #else
+	request = kmalloc(sizeof(*request), GFP_USER);
+#endif
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
 	if (!access_ok(request, sizeof(*request))
+#else
+	if (!access_ok(VERIFY_WRITE, request, sizeof(*request))
 #endif
 		|| __put_user(req32.mode, &request->mode)
 		|| __put_user(req32.buf_width, &request->buf_width)
