@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: GPL-2.0-only
 /*
  * Copyright (C) 2012 Red Hat
- * Copyright (c) 2015 - 2018 DisplayLink (UK) Ltd.
+ * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
  *
  * Based on parts on udlfb.c:
  * Copyright (C) 2009 its respective authors
@@ -12,16 +12,28 @@
  */
 
 #include <linux/version.h>
+#if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_vblank.h>
+#include <drm/drm_damage_helper.h>
+#elif KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+#include <drm/drm_damage_helper.h>
+#else
 #include <drm/drmP.h>
+#endif
 #include <drm/drm_atomic.h>
 #include <drm/drm_crtc.h>
 #include <drm/drm_crtc_helper.h>
 #include <drm/drm_plane_helper.h>
 #include <drm/drm_atomic_helper.h>
 #include "evdi_drm.h"
-#include "evdi_drv.h"
+#include "evdi_drm_drv.h"
 #include "evdi_cursor.h"
 #include "evdi_params.h"
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+#include <drm/drm_gem_atomic_helper.h>
+#else
+#include <drm/drm_gem_framebuffer_helper.h>
+#endif
 
 static void evdi_crtc_dpms(__always_unused struct drm_crtc *crtc,
 			   __always_unused int mode)
@@ -32,6 +44,7 @@ static void evdi_crtc_dpms(__always_unused struct drm_crtc *crtc,
 static void evdi_crtc_disable(__always_unused struct drm_crtc *crtc)
 {
 	EVDI_CHECKPT();
+	drm_crtc_vblank_off(crtc);
 }
 
 static void evdi_crtc_destroy(struct drm_crtc *crtc)
@@ -52,38 +65,44 @@ static void evdi_crtc_set_nofb(__always_unused struct drm_crtc *crtc)
 
 static void evdi_crtc_atomic_flush(
 	struct drm_crtc *crtc
-#if KERNEL_VERSION(4, 3, 0) <= LINUX_VERSION_CODE
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI)
+	, struct drm_atomic_state *state
+#else
 	, __always_unused struct drm_crtc_state *old_state
 #endif
 	)
 {
-	struct drm_crtc_state *state = crtc->state;
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI)
+	struct drm_crtc_state *crtc_state = drm_atomic_get_new_crtc_state(state, crtc);
+#else
+	struct drm_crtc_state *crtc_state = crtc->state;
+#endif
 	struct evdi_device *evdi = crtc->dev->dev_private;
-	unsigned long flags;
+	bool notify_mode_changed = crtc_state->active &&
+				   (crtc_state->mode_changed || evdi_painter_needs_full_modeset(evdi->painter));
+	bool notify_dpms = crtc_state->active_changed || evdi_painter_needs_full_modeset(evdi->painter);
 
-	if (state->event) {
-		spin_lock_irqsave(&crtc->dev->event_lock, flags);
-		drm_crtc_send_vblank_event(crtc, state->event);
-		state->event = NULL;
-		spin_unlock_irqrestore(&crtc->dev->event_lock, flags);
-	}
-	if (state->mode_changed && state->active)
-		evdi_painter_mode_changed_notify(evdi, &state->adjusted_mode);
+	if (notify_mode_changed)
+		evdi_painter_mode_changed_notify(evdi, &crtc_state->adjusted_mode);
 
-	if (state->active_changed)
+	if (notify_dpms)
 		evdi_painter_dpms_notify(evdi,
-			state->active ? DRM_MODE_DPMS_ON : DRM_MODE_DPMS_OFF);
+			crtc_state->active ? DRM_MODE_DPMS_ON : DRM_MODE_DPMS_OFF);
 
-	evdi_painter_send_update_ready_if_needed(evdi);
+	evdi_painter_set_vblank(evdi->painter, crtc, crtc_state->event);
+	evdi_painter_send_update_ready_if_needed(evdi->painter);
+	crtc_state->event = NULL;
 }
 
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE
+#else
 static void evdi_mark_full_screen_dirty(struct evdi_device *evdi)
 {
 	const struct drm_clip_rect rect =
 		evdi_painter_framebuffer_size(evdi->painter);
 
 	evdi_painter_mark_dirty(evdi, &rect);
-	evdi_painter_send_update_ready_if_needed(evdi);
+	evdi_painter_send_update_ready_if_needed(evdi->painter);
 }
 
 static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
@@ -98,7 +117,6 @@ static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
 	struct evdi_device *evdi = dev->dev_private;
 	struct drm_gem_object *obj = NULL;
 	struct evdi_gem_object *eobj = NULL;
-
 	/*
 	 * evdi_crtc_cursor_set is callback function using
 	 * deprecated cursor entry point.
@@ -117,11 +135,7 @@ static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
 	EVDI_CHECKPT();
 	if (handle) {
 		mutex_lock(&dev->struct_mutex);
-#if KERNEL_VERSION(4, 7, 0) > LINUX_VERSION_CODE
-		obj = drm_gem_object_lookup(crtc->dev, file, handle);
-#else
 		obj = drm_gem_object_lookup(file, handle);
-#endif
 		if (obj)
 			eobj = to_evdi_bo(obj);
 		else
@@ -132,21 +146,20 @@ static int evdi_crtc_cursor_set(struct drm_crtc *crtc,
 	evdi_cursor_set(evdi->cursor,
 			eobj, width, height, hot_x, hot_y,
 			format, stride);
-
-#if KERNEL_VERSION(4, 12, 0) <= LINUX_VERSION_CODE
-	drm_gem_object_put_unlocked(obj);
+#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_gem_object_put(obj);
 #else
-	drm_gem_object_unreference_unlocked(obj);
+	drm_gem_object_put_unlocked(obj);
 #endif
 
 	/*
 	 * For now we don't care whether the application wanted the mouse set,
 	 * or not.
 	 */
-	if (evdi_enable_cursor_blending)
-		evdi_mark_full_screen_dirty(evdi);
-	else
+	if (evdi->cursor_events_enabled)
 		evdi_painter_send_cursor_set(evdi->painter, evdi->cursor);
+	else
+		evdi_mark_full_screen_dirty(evdi);
 	return 0;
 }
 
@@ -158,13 +171,14 @@ static int evdi_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
 	EVDI_CHECKPT();
 	evdi_cursor_move(evdi->cursor, x, y);
 
-	if (evdi_enable_cursor_blending)
-		evdi_mark_full_screen_dirty(evdi);
-	else
+	if (evdi->cursor_events_enabled)
 		evdi_painter_send_cursor_move(evdi->painter, evdi->cursor);
+	else
+		evdi_mark_full_screen_dirty(evdi);
 
 	return 0;
 }
+#endif
 
 static struct drm_crtc_helper_funcs evdi_helper_funcs = {
 	.mode_set_nofb  = evdi_crtc_set_nofb,
@@ -175,26 +189,58 @@ static struct drm_crtc_helper_funcs evdi_helper_funcs = {
 	.disable        = evdi_crtc_disable
 };
 
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI)
+static int evdi_enable_vblank(__always_unused struct drm_crtc *crtc)
+{
+	return 1;
+}
+
+static void evdi_disable_vblank(__always_unused struct drm_crtc *crtc)
+{
+}
+#endif
+
 static const struct drm_crtc_funcs evdi_crtc_funcs = {
 	.reset                  = drm_atomic_helper_crtc_reset,
 	.destroy                = evdi_crtc_destroy,
 	.set_config             = drm_atomic_helper_set_config,
 	.page_flip              = drm_atomic_helper_page_flip,
-#if KERNEL_VERSION(4, 14, 0) > LINUX_VERSION_CODE
-	.set_property           = drm_atomic_helper_crtc_set_property,
-#endif
 	.atomic_duplicate_state = drm_atomic_helper_crtc_duplicate_state,
 	.atomic_destroy_state   = drm_atomic_helper_crtc_destroy_state,
 
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE
+#else
 	.cursor_set2            = evdi_crtc_cursor_set,
-	.cursor_move            = evdi_crtc_cursor_move
+	.cursor_move            = evdi_crtc_cursor_move,
+#endif
+#if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(RPI)
+	.enable_vblank          = evdi_enable_vblank,
+	.disable_vblank         = evdi_disable_vblank,
+#endif
 };
 
 static void evdi_plane_atomic_update(struct drm_plane *plane,
-				     struct drm_plane_state *old_state)
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+				     struct drm_atomic_state *atom_state
+#else
+				     struct drm_plane_state *old_state
+#endif
+		)
 {
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	struct drm_plane_state *old_state = drm_atomic_get_old_plane_state(atom_state, plane);
+#else
+#endif
 	struct drm_plane_state *state;
 	struct evdi_device *evdi;
+	struct evdi_painter *painter;
+	struct drm_crtc *crtc;
+
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	struct drm_atomic_helper_damage_iter iter;
+	struct drm_rect rect;
+	struct drm_clip_rect clip_rect;
+#endif
 
 	if (!plane || !plane->state) {
 		EVDI_WARN("Plane state is null\n");
@@ -208,6 +254,8 @@ static void evdi_plane_atomic_update(struct drm_plane *plane,
 
 	state = plane->state;
 	evdi = plane->dev->dev_private;
+	painter = evdi->painter;
+	crtc = state->crtc;
 
 	if (!old_state->crtc && state->crtc)
 		evdi_painter_dpms_notify(evdi, DRM_MODE_DPMS_ON);
@@ -216,19 +264,47 @@ static void evdi_plane_atomic_update(struct drm_plane *plane,
 
 	if (state->fb) {
 		struct drm_framebuffer *fb = state->fb;
+		struct drm_framebuffer *old_fb = old_state->fb;
 		struct evdi_framebuffer *efb = to_evdi_fb(fb);
 
 		const struct drm_clip_rect fullscreen_rect = {
 			0, 0, fb->width, fb->height
 		};
 
-		if (state->fb != old_state->fb ||
-		    evdi_painter_needs_full_modeset(evdi)) {
-			evdi_painter_set_scanout_buffer(evdi, efb);
+		if (!old_fb && crtc)
+			evdi_painter_force_full_modeset(painter);
+
+		if (old_fb &&
+		    fb->format && old_fb->format &&
+		    fb->format->format != old_fb->format->format)
+			evdi_painter_force_full_modeset(painter);
+
+		if (fb != old_fb ||
+		    evdi_painter_needs_full_modeset(painter)) {
+
+			evdi_painter_set_scanout_buffer(painter, efb);
+
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+			state->visible = true;
+			state->src.x1 = 0;
+			state->src.y1 = 0;
+			state->src.x2 = fb->width << 16;
+			state->src.y2 = fb->height << 16;
+
+			drm_atomic_helper_damage_iter_init(&iter, old_state, state);
+			while (drm_atomic_helper_damage_iter_next(&iter, &rect)) {
+				clip_rect.x1 = rect.x1;
+				clip_rect.y1 = rect.y1;
+				clip_rect.x2 = rect.x2;
+				clip_rect.y2 = rect.y2;
+				evdi_painter_mark_dirty(evdi, &clip_rect);
+			}
+#endif
+
+		};
+
+		if (evdi_painter_get_num_dirts(painter) == 0)
 			evdi_painter_mark_dirty(evdi, &fullscreen_rect);
-		} else if (evdi_painter_get_num_dirts(evdi) == 0) {
-			evdi_painter_mark_dirty(evdi, &fullscreen_rect);
-		}
 	}
 }
 
@@ -242,8 +318,17 @@ static void evdi_cursor_atomic_get_rect(struct drm_clip_rect *rect,
 }
 
 static void evdi_cursor_atomic_update(struct drm_plane *plane,
-				      struct drm_plane_state *old_state)
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+				     struct drm_atomic_state *atom_state
+#else
+				     struct drm_plane_state *old_state
+#endif
+		)
 {
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	struct drm_plane_state *old_state = drm_atomic_get_old_plane_state(atom_state, plane);
+#else
+#endif
 	if (plane && plane->state && plane->dev && plane->dev->dev_private) {
 		struct drm_plane_state *state = plane->state;
 		struct evdi_device *evdi = plane->dev->dev_private;
@@ -275,11 +360,7 @@ static void evdi_cursor_atomic_update(struct drm_plane *plane,
 						fb->height,
 						0,
 						0,
-#if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
-						fb->pixel_format,
-#else
 						fb->format->format,
-#endif
 						stride);
 			}
 
@@ -288,7 +369,7 @@ static void evdi_cursor_atomic_update(struct drm_plane *plane,
 		}
 
 		mutex_unlock(&plane->dev->struct_mutex);
-		if (evdi_enable_cursor_blending) {
+		if (!evdi->cursor_events_enabled) {
 			evdi_cursor_atomic_get_rect(&old_rect, old_state);
 			evdi_cursor_atomic_get_rect(&rect, state);
 
@@ -306,11 +387,21 @@ static void evdi_cursor_atomic_update(struct drm_plane *plane,
 }
 
 static const struct drm_plane_helper_funcs evdi_plane_helper_funcs = {
-	.atomic_update = evdi_plane_atomic_update
+	.atomic_update = evdi_plane_atomic_update,
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	.prepare_fb = drm_gem_plane_helper_prepare_fb
+#else
+	.prepare_fb = drm_gem_fb_prepare_fb
+#endif
 };
 
 static const struct drm_plane_helper_funcs evdi_cursor_helper_funcs = {
-	.atomic_update = evdi_cursor_atomic_update
+	.atomic_update = evdi_cursor_atomic_update,
+#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
+	.prepare_fb = drm_gem_plane_helper_prepare_fb
+#else
+	.prepare_fb = drm_gem_fb_prepare_fb
+#endif
 };
 
 static const struct drm_plane_funcs evdi_plane_funcs = {
@@ -336,10 +427,11 @@ static struct drm_plane *evdi_create_plane(
 {
 	struct drm_plane *plane;
 	int ret;
+	char *plane_type = (type == DRM_PLANE_TYPE_CURSOR) ? "cursor" : "primary";
 
 	plane = kzalloc(sizeof(*plane), GFP_KERNEL);
 	if (plane == NULL) {
-		EVDI_ERROR("Failed to allocate primary plane\n");
+		EVDI_ERROR("Failed to allocate %s plane\n", plane_type);
 		return NULL;
 	}
 	plane->format_default = true;
@@ -350,17 +442,13 @@ static struct drm_plane *evdi_create_plane(
 				       &evdi_plane_funcs,
 				       formats,
 				       ARRAY_SIZE(formats),
-#if KERNEL_VERSION(4, 14, 0) <= LINUX_VERSION_CODE
 				       NULL,
-#endif
-				       type
-#if KERNEL_VERSION(4, 5, 0) <= LINUX_VERSION_CODE
-				     , NULL
-#endif
+				       type,
+				       NULL
 				       );
 
 	if (ret) {
-		EVDI_ERROR("Failed to initialize primary plane\n");
+		EVDI_ERROR("Failed to initialize %s plane\n", plane_type);
 		kfree(plane);
 		return NULL;
 	}
@@ -384,12 +472,20 @@ static int evdi_crtc_init(struct drm_device *dev)
 
 	primary_plane = evdi_create_plane(dev, DRM_PLANE_TYPE_PRIMARY,
 					  &evdi_plane_helper_funcs);
+
+#if KERNEL_VERSION(5, 10, 0) <= LINUX_VERSION_CODE
+	cursor_plane = evdi_create_plane(dev, DRM_PLANE_TYPE_CURSOR,
+						&evdi_cursor_helper_funcs);
+#endif
+
+#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE || defined(EL8)
+	drm_plane_enable_fb_damage_clips(primary_plane);
+#endif
+
 	status = drm_crtc_init_with_planes(dev, crtc,
 					   primary_plane, cursor_plane,
-					   &evdi_crtc_funcs
-#if KERNEL_VERSION(4, 5, 0) <= LINUX_VERSION_CODE
-					 , NULL
-#endif
+					   &evdi_crtc_funcs,
+					   NULL
 					   );
 
 	EVDI_DEBUG("drm_crtc_init: %d p%p\n", status, primary_plane);
@@ -398,33 +494,11 @@ static int evdi_crtc_init(struct drm_device *dev)
 	return 0;
 }
 
-static int evdi_atomic_check(struct drm_device *dev,
-			     struct drm_atomic_state *state)
-{
-	struct drm_crtc *crtc;
-	struct drm_crtc_state *crtc_state;
-	int i;
-	struct evdi_device *evdi = dev->dev_private;
-
-	if (evdi_painter_needs_full_modeset(evdi)) {
-#if KERNEL_VERSION(4, 15, 0) <= LINUX_VERSION_CODE
-		for_each_new_crtc_in_state(state, crtc, crtc_state, i) {
-#else
-		for_each_crtc_in_state(state, crtc, crtc_state, i) {
-#endif
-			crtc_state->active_changed = true;
-			crtc_state->mode_changed = true;
-		}
-	}
-
-	return drm_atomic_helper_check(dev, state);
-}
-
 static const struct drm_mode_config_funcs evdi_mode_funcs = {
 	.fb_create = evdi_fb_user_fb_create,
 	.output_poll_changed = NULL,
 	.atomic_commit = drm_atomic_helper_commit,
-	.atomic_check = evdi_atomic_check
+	.atomic_check = drm_atomic_helper_check
 };
 
 void evdi_modeset_init(struct drm_device *dev)
@@ -432,30 +506,20 @@ void evdi_modeset_init(struct drm_device *dev)
 	struct drm_encoder *encoder;
 
 	EVDI_CHECKPT();
+
 	drm_mode_config_init(dev);
 
 	dev->mode_config.min_width = 64;
 	dev->mode_config.min_height = 64;
 
-	dev->mode_config.max_width = 3840;
-	dev->mode_config.max_height = 2160;
+	dev->mode_config.max_width = 7680;
+	dev->mode_config.max_height = 4320;
 
 	dev->mode_config.prefer_shadow = 0;
 	dev->mode_config.preferred_depth = 24;
 
 	dev->mode_config.funcs = &evdi_mode_funcs;
 
-#if KERNEL_VERSION(4, 9, 0) > LINUX_VERSION_CODE
-	drm_mode_create_dirty_info_property(dev);
-#endif
-
-#if KERNEL_VERSION(4, 8, 0) <= LINUX_VERSION_CODE
-
-#elif KERNEL_VERSION(4, 5, 0) <= LINUX_VERSION_CODE
-	drm_dev_set_unique(dev, dev_name(dev->dev));
-#else
-	drm_dev_set_unique(dev, "%s", dev_name(dev->dev));
-#endif
 	evdi_crtc_init(dev);
 
 	encoder = evdi_encoder_init(dev);
