@@ -1,5 +1,5 @@
 /* SPDX-License-Identifier: GPL-2.0-only
- * Copyright (c) 2015 - 2018 DisplayLink (UK) Ltd.
+ * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
  *
  * This file is subject to the terms and conditions of the GNU General Public
  * License v2. See the file COPYING in the main directory of this archive for
@@ -11,6 +11,6 @@
 
 extern unsigned int evdi_loglevel;
 extern unsigned short int evdi_initial_device_count;
-extern bool evdi_enable_cursor_blending;
+extern unsigned short int evdi_disable_texture_import;
 
 #endif /* EVDI_PARAMS_H */
